"""Helpers for parameters processing.

Parameters are typically included in the query string of a URL,
which is the part of a URL that comes after the question mark "?" character.
"""

from vhelpers.types_ import DAny, LParam, TList, UParam


def from_dict(params_d: DAny) -> LParam:
    """Convert a dictionary to a list of parameters.

    :param params_d: A dictionary with keys and values.
    :return:  A list of parameters. If params_d is empty, returns an empty list.

    :example:
        from_dict(params_d={"a": [1, 1]}) -> [("a", 1), ("a", 1)]
    """
    params: LParam = []
    for key, value in params_d.items():
        if isinstance(value, TList):
            for value_ in value:
                params.append((key, value_))
        else:
            params.append((key, value))
    return params


def to_dict(params: UParam) -> DAny:
    """Convert a list of parameters to a dictionary.

    :param params: A list of parameters.
    :return: A dictionary where key is param name.

    :example:
        to_dict(params=[("a", 1), ("a", 1)]) -> {"a": [1, 1]}
    """
    params_d: DAny = {}
    for key, value in params:
        if key in params_d:
            if isinstance(params_d[key], list):
                params_d[key].append(value)
            else:
                params_d[key] = [params_d[key], value]
        else:
            params_d[key] = value
    return params_d



"""Helpers for date processing."""

import datetime
import os

from vhelpers.types_ import LStr, SDate, DInt, Dtime, ODtime

TIME_FMT = "%Y-%m-%d %H:%M:%S"


def aged(delta_: DInt) -> str:
    """Convert time delta to elapsed time in the format %H:%M:%S.

    :param delta_: A dictionary containing the time difference in hours, minutes, and seconds.
    :return: A string representing the elapsed time in the format %H:%M:%S.

    :example:
        aged(hours=8810, minutes=3, seconds=4) -> "8810:03:04"
    """
    hours = delta_["hours"]
    minutes = delta_["minutes"]
    seconds = delta_["seconds"]
    return f"{hours:02}:{minutes:02}:{seconds:02}"


def delta(start: Dtime, end: ODtime = None) -> DInt:
    """Calculate the elapsed days, hours, minutes and seconds between two datetime objects.

    :param start: The starting datetime object.
    :param end: The ending datetime object. If None, the current datetime is used.
    :return: A dictionary containing the elapsed hours, minutes, and seconds.
        If end is None, update data in object.

    :example:
        delta(start="2001-01-02 2:3:4") -> dict(hours=8810, minutes=3, seconds=4)
    """
    if end is None:
        end = datetime.datetime.now()
    delta_ = end - start
    hours, _ = [int(round(t)) for t in divmod(delta_.total_seconds(), 3600)]
    minutes_all, seconds = [int(round(t)) for t in divmod(delta_.total_seconds(), 60)]
    minutes = minutes_all - hours * 60
    return {"hours": hours, "minutes": minutes, "seconds": seconds}


def delta_s(*args, **kwargs) -> str:
    """Calculate the elapsed time in the format %H:%M:%S.

    :param args: The arguments for calculating the time delta.
    :param kwargs: The keyword arguments for calculating the time delta.
    :return: The elapsed time in the format %H:%M:%S.
    """
    time_delta: DInt = delta(*args, **kwargs)
    time_aged = aged(time_delta)
    return time_aged


def last_modified(files: LStr) -> str:
    """Last modified dates."""
    dates: SDate = set()
    for path in files:
        stat = os.stat(path)
        date_ = datetime.datetime.fromtimestamp(stat.st_mtime).date()
        dates.add(date_)
    if not dates:
        return ""
    date_max = max(dates)
    return str(date_max)



"""Helpers for path processing."""

import os
import re
from pathlib import Path

from vhelpers.types_ import LStr, UPath


def get_dirs(root: UPath, pattern: str) -> LStr:
    """Get paths to directories that match required regex pattern in root directory.

    :param root: Root directory to search for files with required pattern.
    :param pattern: Regex pattern to match directory path.
    :return: Paths to directories that match regex pattern.

    :example:
        get_dirs(root="/var/log/", pattern="dir") -> ["/var/log/dir"]
    """
    paths: LStr = []
    for root_i, dirs_i, _ in os.walk(str(root)):
        for dir_ in dirs_i:
            if re.search(pattern, dir_):
                path = os.path.join(root_i, dir_)
                paths.append(path)
    return paths


def get_files(root: UPath, pattern: str) -> LStr:
    """Get paths to files that match required regex pattern in root directory.

    :param root: Root directory to search for files with required pattern.
    :param pattern: Regex pattern to match file path.
    :return: Paths to files that match regex pattern.

    :example:
        get_files(root="/var/log/", pattern="log$") -> ["/var/log/dir/file.log"]
    """
    paths: LStr = []
    for root_i, _, files_i in os.walk(str(root)):
        for file_ in files_i:
            if re.search(pattern, file_):
                path = os.path.join(root_i, file_)
                paths.append(path)
    return paths


def to_dir(path: UPath, name: str) -> Path:
    """Get path to directory with required name.

    :param Path path: Path where need to find directory with required name.
    :param str name: Required directory name.
    :return: Path to directory with required name.
    """
    path_o = Path(path)
    while path_o.name != name:
        if not path_o.name:
            raise ValueError(f"{name=} has not been found in {path!s}")
        path_o = path_o.parent
    return path_o



"""Helpers for YAML processing."""

from vhelpers.types_ import LStr, UStr, LT3Strs, DLStr


def host_cmds(items: LT3Strs) -> str:
    """Create commands in YAML format.

    Where the hostname is the key and the list of commands is the value.
    :param items: List of tuples that contain: hostname, parent command, children commands.
    :return: YAML formatted commands.

    :example:
        items = [("router1", "interface Ethernet1/1", ["description text", "shutdown"])]
        host_cmds(items) -> "
    ---
    host1: |
     cmdA
       cmd1
       cmd2
    "
    """
    hostnames = sorted([t[0] for t in items])
    configs_d: DLStr = {s: [] for s in hostnames}
    for hostname, cmd, cmds in items:
        config = cmd_cmds(cmd=cmd, cmds=cmds)
        configs_d[hostname].append(config)

    yaml_l: LStr = ["---"]
    for hostname, configs in configs_d.items():
        config = "\n".join(configs)
        yaml_l.extend([f"{hostname}: |", config])
    return "\n".join(yaml_l)


def cmd_cmds(cmd: str = "", cmds: UStr = "") -> str:
    """Join parent command and children commands using indentation.

    :param cmd: Parent command.
    :param cmds: Children commands.
    :return: YAML formatted commands with indentation.

    :example:
        cmd_cmds(cmd="cmdA", cmds=["cmd1", "cmd2"]) -> "
     cmdA
       cmd1
       cmd2
    "
    """
    if not cmds:
        return ""
    if isinstance(cmds, str):
        cmds = cmds.split("\n")
    cmds_: LStr = [s.strip() for s in list(cmds)]
    if cmds_ and cmds_[0] == "":
        cmds_ = cmds_[1:]
    if cmds_ and cmds_[-1] == "":
        cmds_ = cmds_[:-1]
    cmds_ = [f"  {s}" for s in cmds_]
    if cmd:
        cmds_.insert(0, f" {cmd.strip()}")
    return "\n".join(cmds_)



"""Helpers for ip addresses processing."""

from ipaddress import IPv4Network

from vhelpers.types_ import LStr, SeqStr


def ip_prefixlen(address: str) -> str:
    """Convert IPv4 address with mask to address with prefix length.

    :param address: IP addresses with mask.
    :return: IP addresses with prefix length.

    :example:
        ip_prefixlen("10.0.0.1 255.255.255.0") -> "10.0.0.1/24"
    """
    ip, mask = address.split()
    network = IPv4Network(f"{ip}/{mask}", strict=False)
    return f"{ip}/{network.prefixlen}"


def ips_prefixlen(addresses: SeqStr) -> LStr:
    """Convert IPv4 addresses with mask to addresses with prefix length.

    :param addresses: A list of IP addresses with mask.
    :return: A list of IP addresses with prefix length.

    :example:
        ips_prefixlen(["10.0.0.1 255.255.255.0"]) -> ["10.0.0.1/24"]
    """
    ips_: LStr = []
    for address in addresses:
        ip = ip_prefixlen(address)
        ips_.append(ip)
    return ips_



"""Helpers for int processing."""

from vhelpers.types_ import IntStr


def to_int(digit: IntStr) -> int:
    """Convert string digit to integer.

    :param digit: Digit, string ot integer.
    :return: Integer or 0 if value is not digit.

    :example:
        to_int(digit="1") -> 1
        to_int(digit="a") -> 0
    """
    if isinstance(digit, int):
        return digit
    digit = digit.strip()
    if not digit.isdigit():
        return 0
    return int(digit)


def to_ordinal(digit: IntStr) -> str:
    """Convert an integer or numeric string to its ordinal representation.

    :param digit: An integer or a string that can be converted to an integer.
    :return: The ordinal string representation of the number.

    :example:
        to_ordinal(digit=1) -> "1st"
        to_ordinal(digit=2) -> "2nd"
        to_ordinal(digit=21) -> "21st"
    """
    int_ = to_int(digit)
    abs_int = abs(int_)

    if int_ == 0:
        return "0th"

    if 10 <= abs_int % 100 <= 20:
        suffix = "th"
    else:
        suffix = {1: "st", 2: "nd", 3: "rd"}.get(abs_int % 10, "th")

    return f"{int_}{suffix}"



"""Typing."""

from datetime import date, datetime
from pathlib import Path
from typing import Any, Dict, List, Optional, Sequence, Set, Tuple, TypeVar, Union

# 1 level
DAny = Dict[str, Any]
DInt = Dict[str, int]
Dtime = datetime
IntStr = Union[int, str]
LAny = List[Any]
LStr = List[str]
ODtime = Optional[datetime]
Param = Tuple[str, Any]
SDate = Set[date]
SStr = Set[str]
SeqStr = Sequence[str]
T = TypeVar("T")
T2Int = Tuple[int, int]
T2Str = Tuple[str, str]
T3Str = Tuple[str, str, str]
T4Str = Tuple[str, str, str, str]
TList = (list, set, tuple)
UPath = Union[Path, str]

# 2 level
DLStr = Dict[str, LStr]
LLAny = List[LAny]
LParam = List[Param]
LT3Str = List[T3Str]
ListTy = List[T]
OLAny = Optional[LAny]
SParam = Set[Param]
SeqTy = Sequence[T]
SetTy = Set[T]
TParam = Tuple[Param, ...]
TupleTy = Tuple[T]
UStr = Union[str, SeqStr]

# 3 level
LListTy = List[ListTy]
T3Strs = Tuple[str, str, UStr]
UParam = Union[LParam, SParam, TParam]
UTy = Union[ListTy, SetTy, TupleTy]

# 4 level
LT3Strs = List[T3Strs]



"""Helpers for regex processing."""

import re

from vhelpers import vint
from vhelpers.types_ import T2Str, T3Str, T4Str, T2Int, SeqStr


def between(
    text: str,
    start: str,
    end: str,
    w_start: bool = False,
    w_end: bool = False,
    strict: bool = False,
) -> str:
    r"""Find all substrings between the start and end regexes.

    :param text: Text where need to find start and end.
    :param start: Regex of start.
    :param end: Regex of end.
    :param w_start: True  - Returns text with matched start text,
                    False - (default) Returns text without matched start text.
    :param w_end: True  - Returns text with matched end text,
                  False - (default) Returns text without matched end text.
    :param strict: True  - Raises ValueError if absent start or end,
                   False - Returns empty string if absent start or end.
    :return: Text between start and end.

    :example:
        between(text="a1\nb2\nc3\nd4", start="b2", end="c3", w_start=True, w_end=True) -> "b2\nc3"
    """
    match_start = re.search(start, text, re.M)
    if not match_start:
        if strict:
            raise ValueError("absent header in text")
        return ""
    header_start = match_start[0]
    idx = match_start.end()
    result_wo_start = text[idx:]

    match_end = re.search(end, result_wo_start, re.M)
    if not match_end:
        if strict:
            raise ValueError("absent footer in text")
        return ""
    header_end = match_end[0]
    idx1 = match_end.start()
    result_wo_end = result_wo_start[:idx1]

    results = [
        header_start if w_start else "",
        result_wo_end,
        header_end if w_end else "",
    ]
    result = "".join(results)
    return result


def find1(pattern: str, string: str, flags: int = 0) -> str:
    """Parse 1 item using findall.

    1 group with parentheses in pattern is required.
    If nothing is found, return 1 empty string.

    :param pattern: The regular expression pattern to search for.
    :param string: The string to search within.
    :param flags: Optional flags to modify the behavior of the search.
    :return: The interested substring, or an empty string if nothing is found.

    :example:
        find1(pattern="a(b)cde", string="abcde") -> "b"
    """
    empty = ""
    result = (re.findall(pattern=pattern, string=string, flags=flags) or [empty])[0]
    if isinstance(result, str):
        return result
    if isinstance(result, tuple):
        return result[0]
    return empty


def find2(pattern: str, string: str, flags: int = 0) -> T2Str:
    """Parse 2 items using findall.

    2 groups with parentheses in pattern is required.
    If nothing is found, return 2 empty strings.

    :param pattern: The regular expression pattern.
    :param string: The string to search within.
    :param flags: Optional flags to modify the behavior of the search.
    :return: A tuple with two interested substrings, or empty strings if nothing is found.

    :example:
        find2(pattern="a(b)(c)de", string="abcde") -> ("b", "c")
    """
    empty = ("", "")
    result = (re.findall(pattern=pattern, string=string, flags=flags) or [empty])[0]
    if isinstance(result, tuple) and len(result) >= 2:
        return result[0], result[1]
    return empty


def find3(pattern: str, string: str, flags: int = 0) -> T3Str:
    """Parse 3 items using findall.

    3 groups with parentheses in pattern is required.
    If nothing is found, returns 3 empty strings.

    :param pattern: The regular expression pattern.
    :param string: The string to search within.
    :param flags: Optional flags to modify the behavior of the search.
    :return: A tuple with three interested substrings, or empty strings if nothing is found.

    :example:
        find3(pattern="a(b)(c)(d)e", string="abcde") -> ("b", "c", "d")
    """
    empty = ("", "", "")
    result = (re.findall(pattern=pattern, string=string, flags=flags) or [empty])[0]
    if isinstance(result, tuple) and len(result) >= 3:
        return result[0], result[1], result[2]
    return empty


def find4(pattern: str, string: str, flags: int = 0) -> T4Str:
    """Parse 4 items using findall.

    4 groups with parentheses in pattern is required.
    If nothing is found, return 4 empty strings.

    :param pattern: The regular expression pattern.
    :param string: The string to search within.
    :param flags: Optional flags to modify the behavior of the search.
    :return: A tuple with three interested substrings, or empty strings if nothing is found.

    :example:
        find4(pattern="a(b)(c)(d)(e)", string="abcde") -> ("b", "c", "d", "e")
    """
    empty = ("", "", "", "")
    result = (re.findall(pattern=pattern, string=string, flags=flags) or [empty])[0]
    if isinstance(result, tuple) and len(result) >= 4:
        return result[0], result[1], result[2], result[3]
    return empty


def find1i(pattern: str, string: str, flags: int = 0) -> int:
    """Parse 1 digit using findall.

    1 group with parentheses in pattern is required.
    If nothing is found, return 0.

    :param pattern: The regular expression pattern to search for.
    :param string: The string to search within.
    :param flags: Optional flags to modify the behavior of the search.
    :return: The interested integer, or 0 if nothing is found.

    :example:
        find1i(pattern="a([0-9]+)b", string="a123b") -> 123
    """
    result = find1(pattern=pattern, string=string, flags=flags)
    return int(result) if result else 0


def find2i(pattern: str, string: str, flags: int = 0) -> T2Int:
    """Parse 2 digits using findall.

    2 groups with parentheses in pattern is required.
    If nothing is found, return tuple of 0.

    :param pattern: The regular expression pattern to search for.
    :param string: The string to search within.
    :param flags: Optional flags to modify the behavior of the search.
    :return: The interested integers, or tuple of 0 if nothing is found.

    :example:
        find2i(pattern="a([0-9])b([0-9])c", string="a1b2c") -> (1, 2)
    """
    result = find2(pattern=pattern, string=string, flags=flags)
    if isinstance(result, tuple) and len(result) >= 2:
        return vint.to_int(result[0]), vint.to_int(result[1])
    return 0, 0


def find1s(patterns: SeqStr, string: str, flags: int = 0) -> str:
    """Parse 1st item that match one of regex in patterns.

    1 group with parentheses in pattern is required.
    If nothing is found, return 1 empty string.

    :param patterns: The list of regular expression patterns to search for.
    :param string: The string to search within.
    :param flags: Optional flags to modify the behavior of the search.
    :return: The interested substring, or an empty string if nothing is found.

    :example:
        find1s(patterns=["a(a)cde", "a(b)cde"], string="abcde") -> "b"
    """
    if not isinstance(patterns, (list, set, tuple)):
        raise TypeError(f"{list} expected")
    for pattern in patterns:
        if result := find1(pattern, string, flags):
            return result
    return ""


def ip(string: str) -> str:
    """Parse 1st IP address from string. If nothing is found, returns an empty string.

    :param string: String where need to find IP address.
    :return: IP address.

    :example:
        ip("text 10.0.0.1/24 10.0.0.2/24 text") -> "10.0.0.1"
    """
    return find1(pattern=r"\d+\.\d+\.\d+\.\d+", string=string)


def prefix(string: str) -> str:
    """Parse 1st prefix from string. If nothing is found, returns an empty string.

    :param string: String where need to find prefix.
    :return: Prefix.

    :example:
        prefix("text 10.0.0.1/24 10.0.0.2/24 text") -> "10.0.0.1/24"
    """
    return find1(pattern=r"\d+\.\d+\.\d+\.\d+/\d+", string=string)



"""Helpers for string processing."""

from vhelpers.types_ import LStr, T2Str


def join(*args) -> str:
    """Join args by delimiter that is first argument, skipping empty strings.

    :param args: Items that need to be joined.
    :return: Joined string.

    :example:
        join(",", "a", "", 0, 1) -> "a,0,1"
    """
    if not args:
        return ""

    delimiter = args[0]
    args = args[1:]

    items: LStr = []
    for item in args:
        if isinstance(item, str):
            if item:
                items.append(item)
        else:
            items.append(str(item))

    return delimiter.join(items)


def join_lines(*args) -> str:
    r"""Join args by new-line character, skipping empty strings.

    :param args: Items that need to be joined.
    :return: Joined string.

    :example:
        join_lines("a", "", 0, 1) -> "a\n0\n1"
    """
    return join("\n", *args)


def repr_info(*args, **kwargs) -> str:
    """Create info without qutes for the __repr__() method.

    :param args: The positional arguments.
    :param kwargs: The keyword arguments.
    :return: A string representation of the parameters.

    :example:
        repr_info("a", "b", c="c", d="d") -> "a, b, c=c, d=d"
    """
    args_ = ", ".join([f"{v!s}" for v in args if v])
    kwargs_ = ", ".join([f"{k}={v!s}" for k, v in kwargs.items() if v])
    params = [s for s in (args_, kwargs_) if s]
    return ", ".join(params)


def repr_params(*args, **kwargs) -> str:
    """Create parameters in quotes for the __repr__() method.

    :param args: The positional arguments.
    :param kwargs: The keyword arguments.
    :return: A string representation of the parameters.

    :example:
        repr_params("a", "b", c="c", d="d") -> "'a', 'b', c='c', d='d'"
    """
    args_ = ", ".join([f"{v!r}" for v in args if v])
    kwargs_ = ", ".join([f"{k}={v!r}" for k, v in kwargs.items() if v])
    params = [s for s in (args_, kwargs_) if s]
    return ", ".join(params)


def reverse(line: str) -> str:
    """Reverse the characters in a string.

    :param line: The input string.
    :return: The reversed string.

    :example:
        reverse("abc") -> "cba"
    """
    return line[::-1]


def split_idx(text: str, idx: int) -> T2Str:
    """Split the text at the specified index.

    :param text: Text to split.
    :param idx: Index at which to split the text.
    :return: Tuple containing the text before the index and the text after the index.

    :example:
        split_idx(text="before_after", idx=7) -> "before_", "after"
    """
    before = text[:idx]
    after = text[idx:]
    return before, after



"""vhelpers."""

from vhelpers import vdate
from vhelpers import vdict
from vhelpers import vint
from vhelpers import vip
from vhelpers import vlist
from vhelpers import vparam
from vhelpers import vpath
from vhelpers import vre
from vhelpers import vstr
from vhelpers import vyml

__all__ = [
    "vdate",
    "vdict",
    "vint",
    "vip",
    "vlist",
    "vparam",
    "vpath",
    "vre",
    "vstr",
    "vyml",
]



"""Helpers for dictionary processing."""

import hashlib
import json
from pathlib import Path
from typing import Any, Type

import tomli

from vhelpers.types_ import UPath, DAny, OLAny


def filter_keys(data: dict, include: OLAny = None, exclude: OLAny = None) -> dict:
    """Filter the data to only required keys by include/exclude parameters.

    :param data: Dictionary to be filtered.
    :param include: Keys that should be present in the filtered dictionary.
    :param exclude: Keys that should not be present in the filtered dictionary.
    :return: New dictionary containing only the required keys.

    :example:
        filter_keys(data={"a": "A", "b": "B"}, include=["a"]) -> {"a": "A"}
        filter_keys(data={"a": "A", "b": "B"}, exclude=["a"]) -> {"b": "B"}
    """
    data_ = data.copy()
    if include:
        data_ = {k: v for k, v in data_.items() if k in include}
    if exclude:
        data_ = {k: v for k, v in data_.items() if k not in exclude}
    return data_


def for_json(data: dict) -> dict:
    """Convert the input data to a format suitable for JSON serialization.

    Replace set with list.

    :param data: Input data to be converted.
    :return: Data converted to a format suitable for JSON serialization.
    :raises TypeError: If input data is not a dictionary.

    :example:
        for_json(data={1: {2, 3}}) -> {1: [2, 3]}
    """
    if not isinstance(data, dict):
        raise TypeError(f"Invalid {type(data)}, expected dict")
    return _for_json(data)


def _for_json(data: Any) -> Any:
    """Convert the input data to a format suitable for JSON serialization.

    Replace set with list.

    :param data: Input data to be converted.
    :return: Data converted to a format suitable for JSON serialization.
    """
    if isinstance(data, set):
        return list(data)  # Convert set to list
    if isinstance(data, dict):
        return {key: _for_json(value) for key, value in data.items()}
    if isinstance(data, list):
        return [_for_json(item) for item in data]
    return data


def invert(data: dict) -> dict:
    """Invert keys and values.

    :param data: Dictionary to invert.
    :return: Dictionary with keys and values inverted.

    :example:
        invert(data={1: 2}) -> {2: 1}
    """
    return {v: k for k, v in data.items()}


def md5hash(data: DAny) -> str:
    """Create MD5 hash of a dictionary.

    :param data: Dictionary to be hashed.
    :return: String representing the MD5 hash of the dictionary.
    """
    dhash = hashlib.md5()
    encoded = json.dumps(data, sort_keys=True).encode()
    dhash.update(encoded)
    return dhash.hexdigest()


def pop(data: dict, key: Any) -> Any:
    """Pop the specified item from the data by key.

    If key is absent in data, do nothing and return None.

    :param data: The dictionary from which the key is to be popped.
    :param key: The key to be popped from the data.
    :return: The popped item if key is present in data, otherwise None.

    :example:
        pop(data={1: 2}, key=1) -> 2
        pop(data={1: 2}, key=3) -> None
    """
    if key in data:
        return data.pop(key)
    return None


def pyproject_d(root: UPath) -> DAny:
    """Convert pyproject.toml to a dictionary.

    :param root: The root directory or path to the pyproject.toml file.
    :return: A dictionary containing the data from pyproject.toml.
    """
    if isinstance(root, str):
        root = Path(root)
    path = Path.joinpath(root, "pyproject.toml")
    with path.open(mode="rb") as file_:
        data = tomli.load(file_)
    return data


def remove_empty(data: dict) -> dict:
    """Remove empty values from a multidimensional dictionary recursively.

    :param data: Dictionary with empty values.
    :return: Dictionary without empty values.

    :example:
        remove_empty(data={1: 1, 2: 0}) -> {1: 1}
    """
    if isinstance(data, dict):
        return {k: remove_empty(v) for k, v in data.items() if v and remove_empty(v) != {}}
    return data


def sha256hash(data: dict) -> int:
    """Create SHA-256 hash of a dictionary.

    :param data: Dictionary to be hashed.
    :return: Integer representing the SHA-256 hash of the dictionary.
    """
    items = tuple(sorted(data.items()))
    json_string = json.dumps(items)
    json_bytes = json_string.encode("utf-8")
    hash_object = hashlib.sha256()
    hash_object.update(json_bytes)
    return int(hash_object.hexdigest(), 16)


# ============================ get typed =============================


def get_any(*args) -> Any:
    """Retrieve a nested value from a dictionary using a sequence of keys.

    :param args: Arguments where the first is the dictionary and the rest are keys
        used to retrieve the desired nested value.
    :return: Value corresponding to the sequence of keys, or None if any key
             is not found or if the access fails.

    :example:
        get_any({"a": {"b": "B"}}, "a", "b") -> "B"
        get_any({"a": {"b": "B"}}, "a", "c") -> None
    """
    try:
        return _get_keys(*args, type_=type(None))
    except (KeyError, IndexError, TypeError):
        return None


def get_bool(*args) -> bool:
    """Retrieve a nested boolean value from a dictionary using a sequence of keys.

    :param args: Arguments where the first is the dictionary and the rest are keys
        used to retrieve the desired nested value.
    :return: Boolean value corresponding to the sequence of keys, or False if any key
             is not found or if the access fails.

    :example:
        get_bool({"a": {"b": True}}, "a", "b") -> True
        get_bool({"a": {"b": "B"}}, "a", "b") -> False
    """
    return _get_keys(*args, type_=bool)


def get_dict(*args) -> dict:
    """Retrieve a nested dictionary value from a dictionary using a sequence of keys.

    :param args: Arguments where the first is the dictionary and the rest are keys
        used to retrieve the desired nested value.
    :return: Dictionary value corresponding to the sequence of keys, or empty dictionary if any key
             is not found or if the access fails.

    :example:
        get_dict({"a": {"b": {"c": "C"}}}, "a", "b") -> {"c": "C"}
        get_dict({"a": {"b": "B"}}, "a", "b") -> {}
    """
    return _get_keys(*args, type_=dict)


def get_int(*args) -> int:
    """Retrieve a nested integer value from a dictionary using a sequence of keys.

    :param args: Arguments where the first is the dictionary and the rest are keys
        used to retrieve the desired nested value.
    :return: Integer value corresponding to the sequence of keys, or 0 if any key
             is not found or if the access fails.

    :example:
        get_int({"a": {"b": "1"}}, "a", "b") -> 1
        get_int({"a": {"b": "B"}}, "a", "b") -> 0
    """
    data, *keys = args
    try:
        for key in keys:
            data = data[key]
    except (KeyError, TypeError):
        return 0

    if isinstance(data, int):
        return data

    if isinstance(data, str) and data.isdigit():
        return int(data)

    return 0


def get_list(*args) -> list:
    """Retrieve a nested list value from a dictionary using a sequence of keys.

    :param args: Arguments where the first is the dictionary and the rest are keys
        used to retrieve the desired nested value.
    :return: List value corresponding to the sequence of keys, or empty list if any key
             is not found or if the access fails.

    :example:
        get_list({"a": {"b": ["B"]}}, "a", "b") -> ["B"]
        get_list({"a": {"b": "B"}}, "a", "b") -> []
    """
    return _get_keys(*args, type_=list)


def get_str(*args) -> str:
    """Retrieve a nested string value from a dictionary using a sequence of keys.

    :param args: Arguments where the first is the dictionary and the rest are keys
        used to retrieve the desired nested value.
    :return: String value corresponding to the sequence of keys, or empty string if any key
             is not found or if the access fails.

    :example:
        get_str({"a": {"b": "B"}}, "a", "b") -> "B"
        get_str({"a": {"b": 1}}, "a", "b") -> ""
    """
    return _get_keys(*args, type_=str)


def _get_keys(*args, type_: Type) -> Any:
    """Retrieve values from data using keys and check their data types.

    :param keys: Chaining dictionary keys to retrieve the desired value.
    :param type_: Data type.
    :param data: Dictionary.

    :return: Value with proper data type.
    """
    data, *keys = args
    try:
        for key in keys:
            data = data[key]  # type: ignore
    except (KeyError, IndexError, TypeError):
        return type_()

    if type_ is type(None):
        return data
    if not isinstance(data, type_):
        return type_()

    return data



"""Helpers for list processing."""

import math
import re
from string import punctuation
from typing import Any, Generator, Sequence

from vhelpers.types_ import SeqTy, ListTy, TList, LStr, LAny, LLAny, LListTy


def dupl(items: SeqTy) -> ListTy:
    """Find duplicates of the items.

    :param items: A list of items where need to find duplicates.
    :return: A list of items with duplicates.

    :example:
        dupl([1, 2, 1]) -> [1]
        dupl([{1}, {2}, {1}]) -> [{1}]
    """
    types = set()
    is_hash_error = False
    seen = set()
    duplicates_hashable = set()
    duplicates = []

    for item in items:
        types.add(type(item))

        if isinstance(item, (int, str)):
            hashable: Any = item
        elif isinstance(item, dict):
            hashable = tuple(item.items())
        elif isinstance(item, set):
            hashable = frozenset(item)
        else:
            try:
                hashable = hash(item)
            except TypeError:
                is_hash_error = True
                hashable = str(item)

        if hashable in seen:
            duplicates_hashable.add(hashable)
            if item not in duplicates:
                duplicates.append(item)
        seen.add(hashable)

    if is_hash_error and len(types) > 1:
        raise TypeError(f"{types=} expected only one.")

    return duplicates


def flatten(items: Sequence) -> list:
    """Convert a multidimensional list to a flattened list.

    :param items: The list to be flattened.
    :return: Flat list.

    :example:
        flatten([1, [2, [3]], 4, [5, [6]]]) -> [1, 2, 3, 4, 5, 6]
    """
    return list(_flatten(items))


def _flatten(items: Sequence, ignore_types=(str, bytes)) -> Generator:
    """Convert a multidimensional list to a flattened generator.

    :param items: The list to be flattened.
    :param ignore_types: Types to be ignored during flattening, defaults to (str, bytes)
    :return: A generator that yields the flattened list.
    """
    for item in items:
        if isinstance(item, Sequence) and not isinstance(item, ignore_types):
            yield from _flatten(item)
        else:
            yield item


def is_in(items1: list, items2: list) -> bool:
    """Check if any item in items1 is present in items2.

    :param items1: A list of items.
    :param items2: A list of items.
    :return: True if any item in items1 is present in items2, False otherwise.

    :example:
        is_in(items1=[1, 2], items2=[2, 3]) is True
        is_in(items1=[1, 2], items2=[2, 3]) is True
    """
    for items in items1:
        if items in items2:
            return True
    return False


def no_dupl(items: SeqTy) -> ListTy:
    """Remove duplicates from a list of items.

    :param items: A list of items.
    :return: A list of items without duplicates.

    :example:
        no_dupl([1, 2, 1]) -> [1, 2]
    """
    items_: ListTy = []
    for item in items:
        if item not in items_:
            items_.append(item)
    return items_


def replace(items: list, old: Any, new: Any) -> None:
    """Replace one item with another.

    :param items: The list of items where need replace item.
    :param old: The item to be replaced.
    :param new: The item to replace with.
    :return: None. Update items.

    :example:
        items = [1, 2, 3]
        replace(items=items, old=2, new=4)
        assert items == [1, 4, 3]
    """
    if old in items:
        idx = items.index(old)
        items[idx] = new


def split(text: str, chars: str = "", ignore: str = "") -> LStr:
    """Split string by punctuation chars.

    :param text: Text to split by punctuation.
    :param chars: Extra punctuation chars.
    :param ignore: Ignore punctuation chars.
    :return: Values without punctuation.

    :example:
        split(text="1; 2_3-4X5,6", chars="_X", ignore=",") -> ["1", "2", "3", "4", "5,6"]
    """
    punctuation_ = punctuation + chars
    for ignore_char in ignore:
        punctuation_ = punctuation_.replace(ignore_char, "")
    punctuation_ = punctuation_.replace("-", "\\-")
    punctuation_ = rf"[\s{punctuation_}]+"
    items: LStr = re.split(punctuation_, text)
    items = [s for s in items if s]
    return items


def to_list(items: Any) -> list:
    """Convert the input items from any into a list.

    If items is a list, set or tuple, simply change its type to list.
    Otherwise, create a list with the value as its first item.
    If items is None return an empty list.

    :param items: The items to be converted into a list.
    :return: The converted list.

    :example:
        to_list((1, 2)) -> [1, 2]
        to_list(1) -> [1]
        to_list(None) -> []
    """
    if items is None:
        return []
    if not isinstance(items, TList):
        return [items]
    return list(items)


def to_lists(items: SeqTy, count: int) -> LListTy:
    """Convert a flat list into a multidimensional list with a fixed number of inner lists.

    :param items: The flat list to convert.
    :param count: The number of inner lists.
    :return: A multidimensional list.

    :example:
        to_lists(items=[1, 2, 3, 4, 5], count=2) -> [[1, 2, 3], [4, 5]]
    """
    count_ = int(count)
    if count_ <= 0:
        return []

    length = int(math.ceil(len(items) / count_))

    items_ = []
    items = list(items)
    while items:
        piece = items[:length]
        items_.append(piece)
        items = items[length:]

    if diff := count - len(items_):
        empty_items: ListTy = [[] for _ in range(diff)]
        items_.extend(empty_items)

    return items_


def to_lstr(items: Any) -> LStr:
    """Convert the input items from any into a list of string.

    If items is a list, set or tuple, simply change its type to list.
    If items is None or empty string return an empty list.

    :param items: The items to be converted into a list of string.
    :return: The converted list.

    :example:
        to_lstr([1, "2"]) -> ["1", "2"]
        to_lstr(1) -> ["1"]
        to_lstr("") -> []
    """
    if items in [None, ""]:
        return []
    if not isinstance(items, TList):
        return [str(items)]
    return [str(i) for i in items]


def to_multi(items: LAny, count: int) -> LLAny:
    """Convert a flat list into a multidimensional list.

    Convert a list with the specified number of items in each inner list.

    :param items: The flat list to convert.
    :param count: The number of items to include in each inner list.
    :return: A multidimensional list with the specified number of items in each inner list.

    :example:
        to_multi(items=[1, 2, 3, 4, 5], count=2) -> [[1, 2], [3, 4], [5]]
    """
    items_ = []
    for i in range(0, len(items), count):
        items_.append(items[i : i + count])
    return items_
